///////////////////////////////////////////////////////////////////////////////
//
// Developed by Gardener of Thoughts (gardenerofthoughts.org).
// Released under a "Common Sense License".
//
///////////////////////////////////////////////////////////////////////////////

using System;
using Fluogen.Dependz;
using Fluogen.Dependz.Color;
using Fluogen.Dependz.Contexts;
using Fluogen.Dependz.FluonGenerators;
using Fluogen.DeployedScripts.Configuration.FluonGenerators;
using Got.Libraries;

// #cfg FluonGenerators\Illumination

namespace Fluogen.DeployedScripts.DeployedScripts.Abstracts.FluonGenerators
{
	/// <summary>
	/// Inspired by the lighting algorithm written by Damien M. Jones for UltraFractal.
	/// </summary>
	public abstract class IlluminatedIteratedFluonGenerator : IteratedFluonGeneratorWithMultiDomain<Illumination, DummyCarryState>
	{
		private double LightX;
		private double LightY;
		private double LightZ;

		public override void InitializeFrameState( IFluonConfiguration fc )
		{
			base.InitializeFrameState( fc );

			double hAngle = SC.HorizontalAngleOfLight.DegreesToRadians();
			double vAngle = SC.VerticalAngleOfLight.DegreesToRadians();

			// Light direction. This is the polar to cartesian form conversion from MDN.
			LightX = hAngle.Cos() * vAngle.Cos();
			LightY = hAngle.Sin() * vAngle.Cos();
			LightZ = vAngle.Sin();
		}

		protected override Mdn[] ProvideDomains( IFluonContext fx, IFluonConfiguration fc, IPixelContext px, Mdn x )
		{
			int axisNumberForWidth = SC.AxisNumberForWidth > 0 ? SC.AxisNumberForWidth : fc.AxisNumberForStratumWidth;
			int axisNumberForHeight = SC.AxisNumberForHeight > 0 ? SC.AxisNumberForHeight : fc.AxisNumberForStratumHeight;

			double wPixelDistance = SC.UseSmallestOrbitSeparation ? px.SmallestDistanceBetweenPixels : px.StratumWidthPixelDistance;
			double hPixelDistance = SC.UseSmallestOrbitSeparation ? px.SmallestDistanceBetweenPixels : px.StratumHeightPixelDistance;

			Mdn xw = x.AddOnAxis( axisNumberForWidth - 1, SC.OrbitSeparationFraction * wPixelDistance );
			Mdn xh = x.AddOnAxis( axisNumberForHeight - 1, SC.OrbitSeparationFraction * hPixelDistance );

			return new Mdn[] { x, xw, xh };
		}

		protected override IGeneratedCodomainContext GetGeneratedCodomainContext( IFluonContext fx, IFluonConfiguration fc, IPixelContext px
			, Mdn x, DummyCarryState carryState, GeneratedCodomainContextForMdn[] generationContexts )
		{
			DetermineSurface( SC, px, generationContexts, out double sx, out double sy, out double sz );

			PerturbateSurface( fc, SC, ref sx, ref sy, ref sz );

			var reducedMdn = ApplyLight( SC, LightX, LightY, LightZ, sx, sy, sz );

			var gc1 = generationContexts[ 0 ];

			// Preserve some "Support", else it's lost completely.
			reducedMdn = reducedMdn.ReplaceSupport( gc1.ReducedMdn.Support );

			return new GeneratedCodomainContextForReducedMdn( x, reducedMdn, gc1.RealizedIterations, gc1.HasEscaped );
		}

		protected virtual void PerturbateSurface( IFluonConfiguration fc, Illumination sc, ref double sx, ref double sy, ref double sz )
		{
			PerturbateSurface( sc, ref sx, ref sy, ref sz );
		}

		protected static void DetermineSurface( Illumination sc, IPixelContext px, GeneratedCodomainContextForMdn[] generationContexts
			, out double sx, out double sy, out double sz )
		{
			var gc1 = generationContexts[ 0 ];
			var gc2 = generationContexts[ 1 ];
			var gc3 = generationContexts[ 2 ];

			double rm1 = gc1.ReducedMdn.Main;
			double rm2 = gc2.ReducedMdn.Main;
			double rm3 = gc3.ReducedMdn.Main;

			double wPixelDistance = sc.UseSmallestOrbitSeparation ? px.SmallestDistanceBetweenPixels : px.StratumWidthPixelDistance;
			double hPixelDistance = sc.UseSmallestOrbitSeparation ? px.SmallestDistanceBetweenPixels : px.StratumHeightPixelDistance;
			double wOrbitSeparation = sc.OrbitSeparationFraction * wPixelDistance;
			double hOrbitSeparation = sc.OrbitSeparationFraction * hPixelDistance;
			double orbitSeparation = MathLib.DiagonalReal( wOrbitSeparation, hOrbitSeparation );

			// Determine the surface height.
			sx = rm2 - rm1;
			sy = rm3 - rm1;
			sz = orbitSeparation;

			// Normalize surface with the 3D radius.
			double sr_3d = MathLib.DiagonalReal( sx, sy, sz );
			sx /= sr_3d;
			sy /= sr_3d;
			sz /= sr_3d;
		}

		protected static void PerturbateSurface( Illumination sc, ref double sx, ref double sy, ref double sz )
		{
			switch( sc.Perturbation )
			{
				case Illumination.SurfacePerturbation.None:
					return;
				case Illumination.SurfacePerturbation.DoubleWhammyX:
					sx = PerturbateSurface( sc, sx );
					return;
				case Illumination.SurfacePerturbation.DoubleWhammyY:
					sy = PerturbateSurface( sc, sy );
					return;
				case Illumination.SurfacePerturbation.DoubleWhammyZ:
					sz = PerturbateSurface( sc, sz );
					return;
				case Illumination.SurfacePerturbation.DoubleWhammyXY:
					sx = PerturbateSurface( sc, sx );
					sy = PerturbateSurface( sc, sy );
					return;
				case Illumination.SurfacePerturbation.DoubleWhammyXZ:
					sx = PerturbateSurface( sc, sx );
					sz = PerturbateSurface( sc, sz );
					return;
				case Illumination.SurfacePerturbation.DoubleWhammyYZ:
					sy = PerturbateSurface( sc, sy );
					sz = PerturbateSurface( sc, sz );
					return;
				case Illumination.SurfacePerturbation.DoubleWhammyXYZ:
					sx = PerturbateSurface( sc, sx );
					sy = PerturbateSurface( sc, sy );
					sz = PerturbateSurface( sc, sz );
					return;
				default:
					throw new InvalidOperationException( $"Unhandled \"{sc.Perturbation}\" surface perturbation." );
			}
		}

		protected static double PerturbateSurface( Illumination sc, double x )
		{
			x = (sc.PerturbationInnerAnchor - x.ApplyTransferFunction( sc.PerturbationInnerTransferFunction ))
				.ApplyTransferFunction( sc.PerturbationOuterTransferFunction );

			x = (sc.PerturbationInnerAnchor - x.ApplyTransferFunction( sc.PerturbationInnerTransferFunction ))
				.ApplyTransferFunction( sc.PerturbationOuterTransferFunction );

			return x;
		}

		protected static ColorDeterminant ApplyLight( Illumination sc, double lx, double ly, double lz, double sx, double sy, double sz )
		{
			// Amount of light on the surface.
			double l = lx * sx + ly * sy + lz * sz;

			if( sc.UseMinimumAmbientLight )
				l = l.Highest( sc.AmbientLight );

			return new ColorDeterminant( l );
		}
	}
}
